# -----------------------------------------------------------------
#    SCRIPT            context.py
#    AUTHOR            Sebastien Paviot
#                    spaviot @ gmail . com
#    DATE:            July,August 2009 - April,May 2010
#                       July 2021
#
#    DESCRIPTION:    Define tool contexts
#
#    VERSION:        2022.0
#
# -----------------------------------------------------------------

from __future__ import print_function
from __future__ import with_statement

import math as math

import maya.OpenMaya as api
import maya.OpenMayaUI as apiui
import maya.cmds as cmds

###################################################

###################################################

spPaint3dContextID = "spPaint3dContext"
spPaint3dTempGroupID = "spPaint3dTempGroup"

# unit conversion dictionary relative to 1 cm (default unit system)
sp3dUnit = {
    "mm": 10,
    "cm": 1,
    "m": 0.01,
    "in": 0.393701,
    "ft": 0.0328084,
    "yd": 0.0109361, }

sp3d_dbgfile = "C:\\sp3ddbg_log.txt"
sp3d_dbg = False  # debug flag to log to file
sp3d_log = False  # debug flag to log to script editor log
sp3d_place = False  # debug flag for place context
sp3d_ramp = False  # debug flag for rampFX
sp3d_MFn = False  # debug flag for MFn stuff


class point(object):
    """
    define an object of 3 attributes used for various purposes
    """

    def __init__(self, x, y, z):
        """
        initialize object attributes
        """
        self.x = x
        self.y = y
        self.z = z

    def asMPoint(self):
        """
        return point as MPoint()
        """
        return api.MPoint(self.x, self.y, self.z)

    def asMFPoint(self):
        """
        return point as MFloatPoint()
        """
        return api.MFloatPoint(self.x, self.y, self.z)

    def asMVector(self):
        """
        return point as MVector()
        """
        return api.MVector(self.x, self.y, self.z)

    def asMFVector(self):
        """
        return point as MFloatVector()
        """
        return api.MFloatVector(self.x, self.y, self.z)


class intersectionPoint(object):
    """
    define an intersection point
    """

    def __init__(self, hitPoint, hitFace, hitTriangle, dagMesh):
        """
        initial setup
        """
        self.valid = None  # bool to validate iteration later on
        self.hitPoint = hitPoint  # point.position tuple
        self.hitFace = hitFace  # face number of the intersection
        self.hitTriangle = hitTriangle  # triangle number in the above face
        self.dagMeshTargetSurface = dagMesh  # MDagPath where the intersection occured (used to avoid having to compute all normals on soon-to-be discarded intersections
        self.timestamp = None  # used to track the creation of an object at that intersectionPoint (later used in the strokePointList class)
        self.dagMeshSourceObject = None  # used to store the DAG path of the created geometry if it's actually a valid intersection
        self.generatedDAG = None  # used to store the DAG path of the created object
        self.initialScale = [1, 1, 1]  # used to store the self.generatedDAG initial scale

    def getHitNormal(self, smooth=False):
        """
        return the normal (MVector) at the self.hitPoint, compute the normal differently according to the smooth boolean argument
        """
        if smooth:  # getting the intersection normal from the MFnMesh method
            normal = api.MVector()
            fnMesh = api.MFnMesh(self.dagMeshTargetSurface)
            fnMesh.getClosestNormal(self.hitPoint.asMPoint(), normal, api.MSpace.kWorld, None)
            return normal

        # compute hard normal (decal mode)
        hitFacept = api.MScriptUtil()
        api.MScriptUtil().setInt(hitFacept.asIntPtr(), self.hitFace)
        itMesh = api.MItMeshPolygon(self.dagMeshTargetSurface)
        itMesh.setIndex(self.hitFace, hitFacept.asIntPtr())
        triVertsArray = api.MPointArray()
        triIndexArray = api.MIntArray()
        itMesh.getTriangle(self.hitTriangle, triVertsArray, triIndexArray, api.MSpace.kWorld)

        return self.getCrossProduct(triVertsArray[0], triVertsArray[1], triVertsArray[2])

    def convertUnit(self, currentUnit):
        """
        will convert the stored intersection internal coordinates into the current unit of the scene
        """
        if currentUnit == 'cm':
            return

        self.hitPoint.x = float(cmds.convertUnit(self.hitPoint.x, fromUnit='cm', toUnit=currentUnit))
        self.hitPoint.y = float(cmds.convertUnit(self.hitPoint.y, fromUnit='cm', toUnit=currentUnit))
        self.hitPoint.z = float(cmds.convertUnit(self.hitPoint.z, fromUnit='cm', toUnit=currentUnit))

    def getCrossProduct(self, p1, p2, p3):
        """
        return the cross product from the 3 points
        """
        vectA = api.MVector((p2.x - p1.x), (p2.y - p1.y), (p2.z - p1.z))
        vectB = api.MVector((p3.x - p1.x), (p3.y - p1.y), (p3.z - p1.z))
        vectA.normalize()
        vectB.normalize()
        return self.doCrossProduct(vectA, vectB)

    def doCrossProduct(self, v1, v2):
        """
        compute the cross product vectors
        """
        vectNorm = api.MVector(((v1.y * v2.z) - (v1.z * v2.y)),
                               ((v1.z * v2.x) - (v1.x * v2.z)),
                               ((v1.x * v2.y) - (v1.y * v2.x)))
        vectNorm.normalize()
        return vectNorm

    def startTimer(self):
        """
        start the timerX for the current object
        """
        self.timestamp = cmds.timerX()

    def updateDAGSourceObject(self, dagstring):
        """
        update self.dagMeshSourceObject with the dagstring argument
        """
        self.dagMeshSourceObject = dagstring

    def createdObjectDAG(self, dagstring):
        """
        update the self.generatedDAG with the dagstring argument
        """
        self.generatedDAG = dagstring

    def setInitialScale(self):
        """
        Set initial scale to the intersected point dag object created
        """
        self.initialScale = cmds.xform(self.generatedDAG, query=True, scale=True, r=True)

    def isValid(self, confirm=None):
        """
        put the valid flag to the passed value. return the current flag state
        """
        if confirm is not None:
            self.valid = confirm
        return self.valid


class intersectionList(object):
    """
    define list of intersection points
    """

    def __init__(self, ipoint=None):
        """
        initial setup
        input: point object, (optional point object)
        """
        self.intersectionList = []
        if ipoint:
            self.addPoint(ipoint)
        if sp3d_log:
            print("creating a new intersectionList (empty? %s)" % self.intersectionList)

    def addPoint(self, ipoint):
        """
        add a point to the list
        """
        self.intersectionList.append(ipoint)
        if sp3d_log:
            print("adding a new intersection to intersectionList (length of list: %i)" % len(self.intersectionList))

    def getLength(self):
        """
        return the length of the sel.intersectionList
        """
        return len(self.intersectionList)

    def getClosest(self, sortorigin):
        """
        will parse the list to return the closest intersectionPoint to the sortorigin argument
        return None if the list is empty
        """
        length = len(self.intersectionList)
        closestintersection = None
        closestdistance = None

        if length == 0:
            return None
        elif length == 1:
            return self.intersectionList[0]

        # list is at least 2 intersections
        for intersection in self.intersectionList:
            if not closestintersection:
                # first intersection considered
                closestintersection = intersection
                closestdistance = getDistanceBetween(intersection.hitPoint, sortorigin)
            else:
                # compare distances
                newdistance = getDistanceBetween(intersection.hitPoint, sortorigin)
                if (newdistance < closestdistance):
                    # new closest
                    closestintersection = intersection
                    closestdistance = newdistance
        # at this point there must be a closesintersetcion
        return closestintersection

    def printList(self):
        """
        debug: print the dag of all objects in the list
        """
        for obj in self.intersectionList:
            print("object created: %s (using source: %s)" % (obj.generatedDAG, obj.dagMeshSourceObject))


class paintContext(object):
    """
    define paintContext
    """

    def __init__(self, uioptions, transformoptions, sourcelist, targetlist):
        """
        initial setup
        """
        # create the tool context
        if cmds.draggerContext(spPaint3dContextID, exists=True):
            cmds.deleteUI(spPaint3dContextID)
        cmds.draggerContext(spPaint3dContextID, pressCommand=self.onPress, dragCommand=self.onDrag,
                            releaseCommand=self.onRelease, name=spPaint3dContextID, cursor='crossHair', undoMode='step')

        # create context local options
        self.runtimeUpdate(uioptions, transformoptions, sourcelist, targetlist)

        # debug purpose
        self.reentrance = 0

        # initialise world up vector
        if (cmds.upAxis(q=True, axis=True)) == "y":
            self.worldUp = api.MVector(0, 1, 0)
        elif (cmds.upAxis(q=True, axis=True)) == "z":
            self.worldUp = api.MVector(0, 0, 1)
        else:
            # can't figure out up vector
            cmds.confirmDialog(title='Weird stuff happening',
                               message='Not getting any proper info on what the current up vector is. Quitting...')
            raise RuntimeError

        # fetch current scene unit
        self.unit = cmds.currentUnit(query=True, linear=True)

    def runContext(self):
        """
        set maya tool to the context
        """
        if cmds.draggerContext(spPaint3dContextID, exists=True):
            cmds.setToolTo(spPaint3dContextID)

    def onPress(self):
        """
        on mouse press initial event
        """
        if sp3d_dbg:
            logDebugInfo('entered paintContext onPress')

        # initialise the intersection list that will contain all the created objects within the same stroke
        self.strokeIntersectionList = intersectionList()

        # create the temporary group used through the stroke to store geometry as they are created
        self.tempgroup = cmds.group(empty=True, name=spPaint3dTempGroupID)

        pressPosition = cmds.draggerContext(spPaint3dContextID, query=True, anchorPoint=True)
        worldPos, worldDir = getViewportClick(pressPosition[0], pressPosition[1])

        intersected = targetSurfaceLoopIntersect(self.targetList, worldPos, worldDir)

        if intersected:
            # there was a usable intersection found

            # first converting the returned internal coordinates into the current unit system
            intersected.convertUnit(self.unit)

            intersected.isValid(True)
            if sp3d_dbg:
                logDebugInfo('found intersected')
            if sp3d_log:
                print('intersection at X: %f | Y: %f | Z: %f' % (
                    intersected.hitPoint.x, intersected.hitPoint.y, intersected.hitPoint.z))
            if not self.uiValues.paintFlux:
                # paintFlux set on timer
                intersected.startTimer()
            # feed various data into the intersectionPoint object
            if self.uiValues.random:
                intersected.dagMeshSourceObject = self.sourceList.getRandom()
            else:
                intersected.dagMeshSourceObject = self.sourceList.getNext()
            if sp3d_dbg:
                logDebugInfo('got the dag for the source object to use')

            if sp3d_dbg:
                logDebugInfo('creating object from the dag')
            intersected.createdObjectDAG(self.createObject(intersected))
            intersected.setInitialScale()
            if sp3d_dbg:
                logDebugInfo('finished creating object from the dag, appending to intersection list')
            self.strokeIntersectionList.addPoint(intersected)

            if self.uiValues.jitter:
                # jittering the created object along the U & V parameters from the UI
                u = self.transform.getRandomJitter('uJitter')
                v = self.transform.getRandomJitter('vJitter')
                yOffset = math.fabs(self.worldUp.y - 1) * v
                zOffset = math.fabs(self.worldUp.z - 1) * v
                cmds.move(u, yOffset, zOffset, intersected.generatedDAG, relative=True)

        if sp3d_dbg:
            logDebugInfo('finished paintContext onPress')
        forceRefresh()

    def onDrag(self):
        """
        on mouse drag event
        """
        # testing something
        if self.reentrance == 0:
            self.reentrance = 1
        else:
            return

        if sp3d_dbg:
            logDebugInfo('entered paintContext onDrag')

        dragPosition = cmds.draggerContext(spPaint3dContextID, query=True, dragPoint=True)

        worldPos, worldDir = getViewportClick(dragPosition[0], dragPosition[1])

        intersected = targetSurfaceLoopIntersect(self.targetList, worldPos, worldDir)
        if intersected:
            # there was a usable intersection found, checking if it's coherent with the paintFlux settings

            # first converting the returned internal coordinates into the current unit system
            intersected.convertUnit(self.unit)

            if len(self.strokeIntersectionList.intersectionList) == 0:
                # there was no intersection during onPress
                if sp3d_log:
                    print('intersection at X: %f | Y: %f | Z: %f' % (
                        intersected.hitPoint.x, intersected.hitPoint.y, intersected.hitPoint.z))
                if not self.uiValues.paintFlux:
                    # paintFlux set on timer
                    intersected.startTimer()

            else:
                # there was at least a previous intersection to test stuff
                if self.uiValues.paintFlux:
                    # paintFlux set on distance
                    distanceToPrevious = getDistanceBetween(self.strokeIntersectionList.intersectionList[-1].hitPoint,
                                                            intersected.hitPoint)
                    correctedPaintDistance = self.uiValues.paintDistance
                    if sp3d_log:
                        print(
                            'intersection at X: %f | Y: %f | Z: %f |||| distance from previous: %f (x: %f | y: %f | z: %f)(threshold: %f)(length of list: %i)' % (
                                intersected.hitPoint.x, intersected.hitPoint.y, intersected.hitPoint.z,
                                getDistanceBetween(self.strokeIntersectionList.intersectionList[-1].hitPoint,
                                                   intersected.hitPoint),
                                self.strokeIntersectionList.intersectionList[-1].hitPoint.x,
                                self.strokeIntersectionList.intersectionList[-1].hitPoint.y,
                                self.strokeIntersectionList.intersectionList[-1].hitPoint.z, correctedPaintDistance,
                                len(self.strokeIntersectionList.intersectionList)))
                    if distanceToPrevious < correctedPaintDistance:
                        # distance with the new intersection is below threshold, forcing an end of proc
                        if sp3d_log:
                            print("breaking from onDrag since < threshold")
                        intersected.isValid(False)
                    else:
                        # distance to previous >= threshold
                        intersected.isValid(True)
                else:
                    # paintFlux set on timer
                    start_time = self.strokeIntersectionList.intersectionList[-1].timestamp
                    if cmds.timerX(startTime=start_time) < self.uiValues.paintTimer:
                        # timer with the new intersection is below threshold, forcing an end of proc
                        intersected.isValid(False)
                    else:
                        # this intersection is valid, starting its timer
                        intersected.isValid(True)
                        intersected.startTimer()

            if intersected.isValid():
                # should only reach here when intersected is above thresholds

                if sp3d_log:
                    print("valid intersection, creating object")
                # feed various data into the intersectionPoint object
                if self.uiValues.random:
                    intersected.updateDAGSourceObject(self.sourceList.getRandom())
                else:
                    intersected.updateDAGSourceObject(self.sourceList.getNext())

                intersected.createdObjectDAG(self.createObject(intersected))
                intersected.setInitialScale()
                self.strokeIntersectionList.addPoint(intersected)

                if self.uiValues.jitter:
                    # jittering the created object along the U & V parameters from the UI
                    u = self.transform.getRandomJitter('uJitter')
                    v = self.transform.getRandomJitter('vJitter')
                    yOffset = math.fabs(self.worldUp.y - 1) * v
                    zOffset = math.fabs(self.worldUp.z - 1) * v
                    cmds.move(u, yOffset, zOffset, intersected.generatedDAG, relative=True)

                if self.uiValues.realTimeRampFX:
                    self.rampFX(self.strokeIntersectionList)

        if sp3d_dbg:
            logDebugInfo('finished paintContext onDrag')
        forceRefresh()
        self.reentrance = 0

    def rampFX(self, objectList):
        """
        operates the ramp FX on the passed intersectionList
        """
        if self.uiValues.rampFX:
            # rampFX is not 0 (1=rotate, 2=scale, 3=both)
            nbObj = objectList.getLength()
            currentObj = 0
            scaleX, scaleY, scaleZ = self.transform.scale
            rotateX, rotateY, rotateZ = self.transform.rotate

            scaleAmplitudeX = scaleX[1] - scaleX[0]
            scaleAmplitudeY = scaleY[1] - scaleY[0]
            scaleAmplitudeZ = scaleZ[1] - scaleZ[0]

            rotateAmplitudeX = rotateX[1] - rotateX[0]
            rotateAmplitudeY = rotateY[1] - rotateY[0]
            rotateAmplitudeZ = rotateZ[1] - rotateZ[0]

            for object in objectList.intersectionList:
                currentObj += 1.0  # to force a float value so the division later on isn't clamped to an integer division

                if self.uiValues.rampFX != 1:
                    currentObjScaleY = currentObjScaleZ = currentObjScaleX = scaleX[0] + scaleAmplitudeX * (
                            currentObj / nbObj)
                    if not self.uiValues.transformScaleUniform:
                        currentObjScaleY = scaleY[0] + scaleAmplitudeY * (currentObj / nbObj)
                        currentObjScaleZ = scaleZ[0] + scaleAmplitudeZ * (currentObj / nbObj)
                    cmds.scale(currentObjScaleX * object.initialScale[0], currentObjScaleY * object.initialScale[1],
                               currentObjScaleZ * object.initialScale[2], object.generatedDAG, relative=False)
                    if sp3d_ramp:
                        print(
                            "rampFX (Scale) obj# %i out of %i (percent: %f)(name: %s) computed scale X %f / Y %f / Z %f (max amplitude X %f)" % (
                                currentObj, nbObj, (currentObj / nbObj), object.generatedDAG, currentObjScaleX,
                                currentObjScaleY, currentObjScaleZ, scaleAmplitudeX))

                if self.uiValues.rampFX != 2:
                    currentObjRotateX = rotateX[0] + rotateAmplitudeX * (currentObj / nbObj)
                    currentObjRotateY = rotateY[0] + rotateAmplitudeY * (currentObj / nbObj)
                    currentObjRotateZ = rotateZ[0] + rotateAmplitudeZ * (currentObj / nbObj)
                    cmds.rotate(currentObjRotateX, currentObjRotateY, currentObjRotateZ, object.generatedDAG,
                                relative=False)

    def onRelease(self):
        """
        on mouse release event: CLEANUP & rampFX if needed
        """
        if not self.uiValues.realTimeRampFX:
            self.rampFX(self.strokeIntersectionList)

        if self.uiValues.hierarchy:
            # grouping objects
            if self.uiValues.group == 0.0:
                # single group sorting
                groupName = self.uiValues.getGroupID()
                for obj in self.strokeIntersectionList.intersectionList:
                    if not cmds.objExists(groupName):
                        groupName = cmds.group(empty=True, name=groupName)
                    cmds.parent(obj.generatedDAG, groupName, relative=True)

            elif self.uiValues.group == 1.0:
                # stroke group sorting
                groupName = cmds.group(empty=True, name='spPaint3dStrokeOutput')
                for obj in self.strokeIntersectionList.intersectionList:
                    cmds.parent(obj.generatedDAG, groupName, relative=True)

            elif self.uiValues.group == 2.0:
                # source group sorting
                for obj in self.strokeIntersectionList.intersectionList:
                    shapeParent = cmds.listRelatives(obj.dagMeshSourceObject, parent=True)
                    groupName = 'spPaint3dOutput_' + shapeParent[0]
                    if not cmds.objExists(groupName):
                        # group doesnt exists, creating
                        groupName = cmds.group(name=groupName, empty=True)

                    cmds.parent(obj.generatedDAG, groupName, relative=True)

        # last cleanup, removing the temp group
        if cmds.objExists(self.tempgroup):
            if sp3d_log:
                print("tempGroup exists, attempting to remove if empty")
            if not cmds.listRelatives(self.tempgroup, children=True):
                # tempGroup is empty, deleting
                if sp3d_log:
                    print("tempGroup (%s) is empty, removing." % self.tempgroup)
                cmds.delete(self.tempgroup)

    def createObject(self, intersection):
        """
        will create the object at the intersection object gathered data, pending all ui and transform options
        will update the stored data to store the created object DAG path and return the newly created object DAG Path back
        """

        newObjectDAG = None
        if self.uiValues.instance:
            if sp3d_dbg:
                logDebugInfo('creating instance')
            # fetching the transform for that shape (instance dont create object of child objects if it's the shape that gets instanced)
            tempDAG = cmds.listRelatives(intersection.dagMeshSourceObject, parent=True)
            newObjectDAG = cmds.instance(tempDAG[0])
        else:
            # fetching the parent transform to prevent some issue while duplicating with preserve input connections and stuff docked onto the transform and not the shape
            if sp3d_dbg:
                logDebugInfo('duplicating object')
            tempDAG = cmds.listRelatives(intersection.dagMeshSourceObject, parent=True)
            newObjectDAG = cmds.duplicate(tempDAG[0], ic=self.uiValues.preserveConn)
        if sp3d_dbg:
            logDebugInfo('DONE creating instance / duplicating object')

        if len(newObjectDAG) > 1:
            print('warning: multiple objects created')

        moveTo(newObjectDAG[0], intersection.hitPoint)

        if self.uiValues.align:
            # align to surface normal
            if sp3d_dbg:
                logDebugInfo('aligning object with surface normal')
            rx, ry, rz = getEulerRotationQuaternion(self.worldUp, intersection.getHitNormal(self.uiValues.smoothNormal))
            cmds.xform(newObjectDAG[0], ro=(rx, ry, rz))
            if sp3d_dbg:
                logDebugInfo('DONE aligning object with surface normal')

        if self.uiValues.transformRotate and not self.uiValues.rampFX:
            # rotate transform
            randrotate = self.transform.getRandomRotate()
            cmds.rotate(randrotate[0], randrotate[1], randrotate[2], newObjectDAG[0], os=True, r=True, rotateXYZ=True)

        if self.uiValues.transformScale and not self.uiValues.rampFX:
            # scale transform
            randscale = self.transform.getRandomScale(self.uiValues.transformScaleUniform)
            cmds.scale(randscale[0], randscale[1], randscale[2], newObjectDAG[0], relative=True)

        if self.uiValues.upOffset != 0:
            offsetArray = [self.uiValues.upOffset * self.worldUp.x,
                           self.uiValues.upOffset * self.worldUp.y,
                           self.uiValues.upOffset * self.worldUp.z]
            cmds.move(offsetArray[0], offsetArray[1], offsetArray[2], newObjectDAG[0], relative=True)

        if self.uiValues.hierarchy:
            # grouping objects
            grouped = None
            if cmds.objExists(self.tempgroup):
                grouped = cmds.parent(newObjectDAG[0], self.tempgroup, relative=True)
            else:
                print('error: temp group does not exist')
            return grouped[0]
        else:
            return newObjectDAG[0]

    def runtimeUpdate(self, uioptions, transformoptions, sourcelist, targetlist):
        """
        entry method used from GUI to pass changes in the UI options through self.ctx.runtimeUpdate(...)
        """
        self.uiValues = uioptions
        self.transform = transformoptions
        self.sourceList = sourcelist
        self.targetList = targetlist


class modifierManager(object):
    """
    Wrapper to manage the modifier keypress / release used in the place context
    """
    modifierMask = {
        'shift': 1,
        'ctrl': 4,
        'alt': 8
    }

    def __init__(self):
        """
        setup various variables used
        """
        self.ctrlReleased = True
        self.shiftReleased = True
        self.altReleased = True

    def resetCtrl(self):
        """
        reset ctrl back to default settings, usually after the key press was acted upon
        """
        pass

    def getState(self):
        """
        return 3 booleans for ctrl / shift / alt state keypress event
        """
        ctrl = shift = alt = False
        modifiers = cmds.getModifiers()

        # SHIFT EVENT
        if (modifiers & 1) > 0:
            if self.shiftReleased:
                # shift was released in the previous iteration, this is a keypress
                self.shiftReleased = False
                shift = True
        else:
            # shift is not currently pressed, resetting the tracking self variable to released state
            self.shiftReleased = True

        # CTRL EVENT
        if (modifiers & 4) > 0:
            if self.ctrlReleased:
                # ctrl was released in the previous iteration, this is a keypress
                self.ctrlReleased = False
                ctrl = True
        else:
            # ctrl is not currently pressed, resetting the tracking self variable to released state
            self.ctrlReleased = True

        # ALT EVENT
        if (modifiers & 8) > 0:
            if self.altReleased:
                # alt was released in the previous iteration, this is a keypress
                self.altReleased = False
                alt = True
        else:
            # alt is not currently pressed, resetting the tracking self variable to released state
            self.altReleased = True

        return ctrl, shift, alt

    def isPressed(self, modifier):
        """
        returns True if the modifier is currently pressed
        """
        bitmask = cmds.getModifiers() & self.modifierMask[modifier]
        return bitmask > 0


class placeCursor(object):
    """
    define a cursor object for use with placeContext
    """

    def __init__(self, sourcedag=None, cursordag=None):
        """
        initialize variables
        """
        self.position = None
        self.rotation = None
        self.rotationIncrement = None
        self.initialScale = [1, 1, 1]
        self.cursorAlign = None
        if sourcedag and cursordag:
            self.setCursorDAG(sourcedag, cursordag)

    def setCursorDAG(self, sourcedag, cursordag, deleteprevious=False):
        """
        update the cursor dag
        """
        self.sourceDAG = sourcedag
        self.sourceDAGPos = getPosition(sourcedag)
        if deleteprevious and cmds.objExists(self.cursorDAG):
            # deleting the previous cursor object and its parent group if it's empty
            parentgroup = cmds.listRelatives(self.cursorDAG, parent=True)
            cmds.delete(self.cursorDAG)
            if not cmds.listRelatives(parentgroup):
                cmds.delete(parentgroup)
        self.cursorDAG = cursordag
        self.initialScale = cmds.xform(self.cursorDAG, query=True, scale=True, r=True)

    def setCursorTransform(self, rotate, scale):
        """
        store transform tuples for rotate and scale of the cursor
        """
        self.rotate = rotate
        self.scale = scale

    def move(self, position=None, rotation=None):
        """
        move the cursor object to position of point type
        """
        if not position:
            # resetting current position to previously stored position if any
            if self.position:
                moveTo(self.cursorDAG, self.position)
                if rotation:
                    cmds.rotate(self.rotationIncrement[0], self.rotationIncrement[1], self.rotationIncrement[2],
                                self.cursorDAG, os=True, r=True, rotateXYZ=True)
        else:
            self.position = position
            moveTo(self.cursorDAG, position)
            if rotation:
                cmds.rotate(self.rotationIncrement[0], self.rotationIncrement[1], self.rotationIncrement[2],
                            self.cursorDAG, os=True, r=True, rotateXYZ=True)

    def rotateCursor(self, increment):
        """
        rotate the cursor object
        """
        if self.rotationIncrement:
            self.rotationIncrement = [(self.rotationIncrement[0] + increment[0]),
                                      (self.rotationIncrement[1] + increment[1]),
                                      (self.rotationIncrement[2] + increment[2])]
        else:
            self.rotationIncrement = increment

        # rotating the cursor from the rotate increment
        cmds.rotate(increment[0], increment[1], increment[2], self.cursorDAG, os=True, r=True, rotateXYZ=True)

    def align(self, rx=None, ry=None, rz=None):
        """
        orient the cursor along the supplied rotation angles
        """
        if rx is not None and ry is not None and rz is not None:
            self.cursorAlign = rx, ry, rz
            cmds.xform(self.cursorDAG, ro=(rx, ry, rz))
        else:
            # realigning cursor with previously stored rotate tuple if it exists
            if self.cursorAlign:
                rx, ry, rz = self.cursorAlign
                cmds.xform(self.cursorDAG, ro=(rx, ry, rz))

    def transform(self, rotate=False, scale=False):
        """
        transform the cursor with self transform scale and rotate
        """
        if rotate:
            cmds.rotate(self.rotate[0], self.rotate[1], self.rotate[2], self.cursorDAG, os=True, r=True, rotateXYZ=True)
        if scale:
            cmds.scale(self.scale[0], self.scale[1], self.scale[2], self.cursorDAG, relative=False)

    def asTemplate(self, mode=True):
        """
        set cursor object in or out of template mode
        """
        # TODO: Check if cursorDAG is a transform
        if mode:
            # set to template
            cmds.setAttr(self.cursorDAG + '.overrideEnabled', 1)
            cmds.setAttr(self.cursorDAG + '.overrideDisplayType', 1)
        else:
            # set to normal
            cmds.setAttr(self.cursorDAG + '.overrideEnabled', 0)
            cmds.setAttr(self.cursorDAG + '.overrideDisplayType', 0)


class placeContext(object):
    """
    define placeContext
    """

    def __init__(self, uioptions, transformoptions, sourcelist, targetlist):
        """
        initial setup
        """
        # create the tool context
        if cmds.draggerContext(spPaint3dContextID, exists=True):
            cmds.deleteUI(spPaint3dContextID)
        cmds.draggerContext(spPaint3dContextID, pressCommand=self.onPress, prePressCommand=self.onBeforePress,
                            dragCommand=self.onDrag, holdCommand=self.onHold, releaseCommand=self.onRelease,
                            name=spPaint3dContextID, cursor='crossHair', undoMode='step')

        # create context local options
        self.runtimeUpdate(uioptions, transformoptions, sourcelist, targetlist)

        # initialise world up vector
        if (cmds.upAxis(q=True, axis=True)) == "y":
            self.worldUp = api.MVector(0, 1, 0)
        elif (cmds.upAxis(q=True, axis=True)) == "z":
            self.worldUp = api.MVector(0, 0, 1)

        # fetch current scene unit
        self.unit = cmds.currentUnit(query=True, linear=True)

        self.reentrance = 0
        self.mState = modifierManager()

    def runContext(self):
        """
        set maya tool to the context
        """
        if cmds.draggerContext(spPaint3dContextID, exists=True):
            cmds.setToolTo(spPaint3dContextID)

    def fetchCursorObject(self):
        """
        gather necessary data and generate a new cursor Object and return the string to the source DAG and the cursor DAG
        """
        # fetch a new cursorObject
        if self.uiValues.random:
            sourceDAG = self.sourceList.getRandom()
        else:
            sourceDAG = self.sourceList.getNext()

        # checking if the sourceDAG is a transform, if not attempting to find the parent transform
        # instancing multiple object doesnt work if using a shape
        # duplicating input connections on the parent transform is not carried over if using the shape
        if cmds.nodeType(sourceDAG) != 'transform':
            tempDAG = cmds.listRelatives(sourceDAG, parent=True)
            sourceDAG = tempDAG[0]

        if self.uiValues.instance:
            newObjectDAG = cmds.instance(sourceDAG)
        else:
            newObjectDAG = cmds.duplicate(sourceDAG, ic=self.uiValues.preserveConn)

        if len(newObjectDAG) > 1:
            print('warning: multiple objects created')

        # create the temporary group used through the stroke to store geometry as they are created
        self.tempgroup = cmds.group(empty=True, name=spPaint3dTempGroupID)
        newObjectDAG = cmds.parent(newObjectDAG[0], self.tempgroup, relative=True)

        return sourceDAG, newObjectDAG[0]

    def fetchCursorTransform(self):
        """
        compute transform data for cursor object and return the tuple for rotate and scale
        """
        # getting the proper transform tuples
        if self.uiValues.transformRotate:
            cursorRotate = self.transform.getRandomRotate()
        else:
            # transform rotate off
            cursorRotate = (0, 0, 0)

        if self.uiValues.transformScale:
            tempCursorScale = self.transform.getRandomScale(self.uiValues.transformScaleUniform)
            cursorScale = [tempCursorScale[0] * self.cursor.initialScale[0],
                           tempCursorScale[1] * self.cursor.initialScale[1],
                           tempCursorScale[2] * self.cursor.initialScale[2]]
        else:
            # transform scale off
            cursorScale = (1, 1, 1)

        return cursorRotate, cursorScale

    def ctrlEvent(self):
        """
        ctrl event stuff
        """
        # fetching a new cursor
        newSourceDAG, newCursorDAG = self.fetchCursorObject()
        self.cursor.setCursorDAG(newSourceDAG, newCursorDAG, True)  # flagging to delete previous cursor
        if self.uiValues.align:
            self.cursor.align()
        self.cursor.move(None,
                         self.cursor.rotationIncrement)  # defaulting without parameters to position the new cursor to the previously stored position
        self.cursor.transform(self.uiValues.transformRotate, self.uiValues.transformScale)
        if self.uiValues.upOffset != 0:
            offsetArray = [self.uiValues.upOffset * self.worldUp.x,
                           self.uiValues.upOffset * self.worldUp.y,
                           self.uiValues.upOffset * self.worldUp.z]
            cmds.move(offsetArray[0], offsetArray[1], offsetArray[2], self.cursor.cursorDAG, relative=True)

    def shiftEvent(self):
        """
        shift event stuff (rotate on upvector)
        """
        rotateArray = [self.worldUp.x * self.uiValues.placeRotate,
                       self.worldUp.y * self.uiValues.placeRotate,
                       self.worldUp.z * self.uiValues.placeRotate]
        self.cursor.rotateCursor(rotateArray)
        if sp3d_log:
            print(self.cursor.rotationIncrement)

    def onBeforePress(self):
        """
        prePress event to setup the temp data for the cursor object
        """
        sourceDAG, cursorDAG = self.fetchCursorObject()
        self.cursor = placeCursor(sourceDAG, cursorDAG)

        cursorRotate, cursorScale = self.fetchCursorTransform()
        self.cursor.setCursorTransform(cursorRotate, cursorScale)
        self.cursor.transform(self.uiValues.transformRotate, self.uiValues.transformScale)
        # self.cursor.asTemplate() #issues with place and object hierarchies, breaking the method for some reason

    def onPress(self):
        """
        on mouse press initial event
        """
        pressPosition = cmds.draggerContext(spPaint3dContextID, query=True, anchorPoint=True)

        # initializing / reseting the rotation increment if we are re-entering place
        self.cursor.rotationIncrement = 0

        ctrl, shift, alt = self.mState.getState()

        worldPos, worldDir = getViewportClick(pressPosition[0], pressPosition[1])

        intersected = targetSurfaceLoopIntersect(self.targetList, worldPos, worldDir)
        if intersected:
            # there was a usable intersection found
            # first checking and converting units if necessary
            intersected.convertUnit(self.unit)
            # now moving the cursor
            self.cursor.move(intersected.hitPoint)
            if self.uiValues.align:
                rx, ry, rz = getEulerRotationQuaternion(self.worldUp,
                                                        intersected.getHitNormal(self.uiValues.smoothNormal))
                self.cursor.align(rx, ry, rz)

            # retransforming the cursor rotation if necessary
            self.cursor.transform(self.uiValues.transformRotate)

            if self.uiValues.upOffset != 0:
                offsetArray = [self.uiValues.upOffset * self.worldUp.x,
                               self.uiValues.upOffset * self.worldUp.y,
                               self.uiValues.upOffset * self.worldUp.z]
                cmds.move(offsetArray[0], offsetArray[1], offsetArray[2], self.cursor.cursorDAG, relative=True)

        else:
            pass
            # no intersection found
            # TODO
            # moving cursor to worldPos and aligning to worldDir

        # forcing a viewport redraw
        forceRefresh()

    def onDrag(self):
        """
        on mouse drag event
        """
        if self.reentrance == 1:
            return
        self.reentrance = 1

        dragPosition = cmds.draggerContext(spPaint3dContextID, query=True, dragPoint=True)

        worldPos, worldDir = getViewportClick(dragPosition[0], dragPosition[1])

        ctrl, shift, alt = self.mState.getState()
        #
        # TODO scale and rotate depending on mouse drag direction
        #
        intersected = targetSurfaceLoopIntersect(self.targetList, worldPos, worldDir)
        if intersected:
            # there was a usable intersection found
            # first checking and converting units if necessary
            intersected.convertUnit(self.unit)
            # now moving the cursor
            if self.uiValues.align:
                rx, ry, rz = getEulerRotationQuaternion(self.worldUp,
                                                        intersected.getHitNormal(self.uiValues.smoothNormal))
                self.cursor.align(rx, ry, rz)
            self.cursor.move(intersected.hitPoint, self.cursor.rotationIncrement)

            if self.uiValues.transformRotate or self.uiValues.transformScale:
                # retransforming the cursor rotation if necessary
                self.cursor.transform(self.uiValues.transformRotate)

                if self.uiValues.continuousTransform:
                    cursorRotate, cursorScale = self.fetchCursorTransform()
                    self.cursor.setCursorTransform(cursorRotate, cursorScale)
                    self.cursor.transform(self.uiValues.transformRotate, self.uiValues.transformScale)

            if self.uiValues.upOffset != 0:
                offsetArray = [self.uiValues.upOffset * self.worldUp.x,
                               self.uiValues.upOffset * self.worldUp.y,
                               self.uiValues.upOffset * self.worldUp.z]
                cmds.move(offsetArray[0], offsetArray[1], offsetArray[2], self.cursor.cursorDAG, relative=True)

        else:
            pass
            # no intersection found
            # TODO
            # moving cursor to worldPos and aligning to worldDir

        if ctrl:
            self.ctrlEvent()
        if shift:
            self.shiftEvent()

        forceRefresh()
        self.reentrance = 0

    def onHold(self):
        """
        on mouse hold event
        """
        if self.reentrance == 1:
            return
        self.reentrance = 1

        dragPosition = cmds.draggerContext(spPaint3dContextID, query=True, dragPoint=True)

        ctrl, shift, alt = self.mState.getState()

        if ctrl:
            self.ctrlEvent()
        if shift:
            self.shiftEvent()

        forceRefresh()

        if sp3d_log:
            message = 'key press detected: '
            if ctrl:
                message += 'ctrl pressed... '
            if shift:
                message += 'shift pressed... '
            if alt:
                message += 'alt pressed... '
            print(message)

        self.reentrance = 0

    def onRelease(self):
        """
        on mouse release event: CLEANUP
        """
        # reverting template
        # self.cursor.asTemplate(False) #issues with object hierarchies for some reason

        if self.uiValues.hierarchy:
            # grouping object
            # parenting the cursor object in the appropriate groupe
            if self.uiValues.group == 0.0:
                # single group sorting
                groupName = self.uiValues.getGroupID()
                if not cmds.objExists(groupName):
                    groupName = cmds.group(empty=True, name=groupName)
                cmds.parent(self.cursor.cursorDAG, groupName, relative=True)

            elif self.uiValues.group == 1.0:
                # stroke group sorting
                groupName = cmds.group(empty=True, name='spPaint3dStrokeOutput')
                cmds.parent(self.cursor.cursorDAG, groupName, relative=True)

            elif self.uiValues.group == 2.0:
                # source group sorting
                shapeParent = cmds.listRelatives(self.cursor.sourceDAG, parent=True)
                groupName = 'spPaint3dOutput_' + shapeParent[0]
                if not cmds.objExists(groupName):
                    # group doesnt exists, creating
                    groupName = cmds.group(name=groupName, empty=True)

                cmds.parent(self.cursor.cursorDAG, groupName, relative=True)

        # last cleanup, removing the temp group
        if cmds.objExists(self.tempgroup) and not cmds.listRelatives(self.tempgroup, children=True):
            # tempGroup is empty, deleting
            cmds.delete(self.tempgroup)

    def runtimeUpdate(self, uioptions, transformoptions, sourcelist, targetlist):
        """
        entry method used from GUI to pass changes in the UI options through self.ctx.runtimeUpdate(...)
        """
        self.uiValues = uioptions
        self.transform = transformoptions
        self.sourceList = sourcelist
        self.targetList = targetlist


# -------------------------------
# Misc Utils
# -------------------------------

def forceRefresh():
    """
    force a current viewport refresh
    """
    cmds.refresh(cv=True)


def moveTo(dag, pos, rot=None):
    """
    move the dag object to pos position
    attemps to compensate for unfrozen transform by reading the scalepivot of the object
    """
    scalePivot = cmds.xform(dag, query=True, ws=True, sp=True)
    transform = cmds.xform(dag, query=True, ws=True, t=True)

    cmds.xform(dag, t=((transform[0] - scalePivot[0]) + pos.x,
                       (transform[1] - scalePivot[1]) + pos.y,
                       (transform[2] - scalePivot[2]) + pos.z))
    if rot:
        if sp3d_log:
            print(rot)
        cmds.rotate(rot[0], rot[1], rot[2], dag, os=True, r=True, rotateXYZ=True)


def getPosition(dag):
    """
    retrieve the world position of the dag parameter object and return a point object containing the position
    """
    tempdag = dag
    if cmds.nodeType(tempdag) != 'transform':
        # get the transform to that shape
        temprelatives = cmds.listRelatives(tempdag, parent=True)
        tempdag = temprelatives[0]

    scalePivot = cmds.xform(tempdag, query=True, ws=True, sp=True)
    transform = cmds.xform(tempdag, query=True, ws=True, t=True)

    return point((transform[0] - scalePivot[0]), (transform[1] - scalePivot[1]), (transform[2] - scalePivot[2]))


def getEulerRotationQuaternion(upvector, directionvector):
    """
    returns the x,y,z degree angle rotation corresponding to a direction vector
    input: upvector (MVector) & directionvector (MVector)
    """
    quat = api.MQuaternion(upvector, directionvector)
    quatAsEuler = quat.asEulerRotation()

    return math.degrees(quatAsEuler.x), math.degrees(quatAsEuler.y), math.degrees(quatAsEuler.z)


def getViewportClick(screenX, screenY):
    """
    return world position and direction of the viewport clicked point (returns point objects)
    """
    maya3DViewHandle = apiui.M3dView()
    activeView = maya3DViewHandle.active3dView()

    clickPos = api.MPoint()
    clickDir = api.MVector()

    activeView.viewToWorld(int(screenX), int(screenY), clickPos, clickDir)

    worldPos = point(clickPos.x, clickPos.y, clickPos.z)
    worldDir = point(clickDir.x, clickDir.y, clickDir.z)

    return worldPos, worldDir


def getCameraFarClip():
    """
    Return current camera far clip
    """
    maya3DViewHandle = apiui.M3dView()
    activeView = maya3DViewHandle.active3dView()

    cameraDP = api.MDagPath()
    maya3DViewHandle.active3dView().getCamera(cameraDP)

    camFn = api.MFnCamera(cameraDP)
    return camFn.farClippingPlane()


def targetSurfaceLoopIntersect(targetList, clickPos, clickDir):
    """
    loop through all the object in targetList and intersect them with click (world pos, direction). creates an intersectionPoint object for each intersection
    sort the list of intersection and return the closest intersectionPoint object from the click world position, return None if no intersection found
    """
    ilist = intersectionList()
    farclip = getCameraFarClip()
    for obj in targetList.obj.keys():
        data = targetList.obj[obj]
        # loop through each object in the targetList
        intersected = intersectTargetSurface(data[0], clickPos, clickDir, farclip)
        if intersected:
            # got meh an intersected
            ilist.addPoint(intersected)

    return ilist.getClosest(clickPos)


def intersectTargetSurface(targetdag, clickPos, clickDir, farclip=1.0):
    """
    intersect a single object from the click world pos and direction. optional farclip distance
    return an intersectionPoint object if there was any intersection
    return None otherwise
    """
    currentHitFP = api.MFloatPoint()  # current intersection
    hitFace = api.MScriptUtil()
    hitTri = api.MScriptUtil()

    hitFace.createFromInt(0)
    hitTri.createFromInt(0)

    hitFaceptr = hitFace.asIntPtr()
    hitTriptr = hitTri.asIntPtr()

    targetDAGPath = getDAGObject(targetdag)

    if targetDAGPath:
        # returned targetDAGPath is sort of valid
        fnMesh = api.MFnMesh(targetDAGPath)
        hit = fnMesh.closestIntersection(clickPos.asMFPoint(),
                                         clickDir.asMFVector(),
                                         None,
                                         None,
                                         True,
                                         api.MSpace.kWorld,
                                         farclip,
                                         True,
                                         None,
                                         currentHitFP,
                                         None,
                                         hitFaceptr,
                                         hitTriptr,
                                         None,
                                         None)
        if hit:
            # there was a positive intersection
            if sp3d_MFn:
                print("Face Hit: %i || Tri Hit: %i" % (hitFace.getInt(hitFaceptr), hitTri.getInt(hitTriptr)))
            return intersectionPoint(point(currentHitFP.x, currentHitFP.y, currentHitFP.z),
                                     hitFace.getInt(hitFaceptr),
                                     hitTri.getInt(hitTriptr),
                                     targetDAGPath)

    # reaches here only if no intersection or not a valid targetDAGPath
    return None


def getDAGObject(dagstring):
    """
    return the DAG Api object from the dagstring argument
    return None if the minimum checks on dagstring don't checkout
    """
    sList = api.MSelectionList()
    meshDP = api.MDagPath()
    api.MGlobal.getSelectionListByName(dagstring, sList)
    sList.getDagPath(0, meshDP)

    return meshDP


def getDistanceBetween(source, target):
    """
    return the distance between source and target
    """
    distance = math.sqrt(math.pow((source.x - target.x), 2) +
                         math.pow((source.y - target.y), 2) +
                         math.pow((source.z - target.z), 2))
    return distance


def getCorrectedDistance(distance, unit):
    """
    return the corrected distance using proper unit to convert back to centimeters
    """
    if unit == 'cm':
        return distance

    return (distance / (sp3dUnit[unit]))


def logDebugInfo(info):
    """
    overwrite the default debug file content with <info>
    """
    with open(sp3d_dbgfile, 'w') as f:
        f.write(info)
